"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple

from mipa.ext.commands._types import _BaseCommand
from mipa.ext.commands.core import Command

if TYPE_CHECKING:
    from mipa.ext.commands.bot import BotBase


class CogMeta(type):
    __cog_name__: str

    def __new__(cls, *args: Tuple[Any], **kwargs: Dict[str, Any]):
        name, bases, attrs = args
        attrs['__cog_name__'] = kwargs.pop('name', name)
        attrs['__cog_settings__'] = kwargs.pop('command_attrs', {})
        listeners = {}
        commands = {}
        no_bot_cog = 'Commands or listeners must not start with cog_ or bot_ (in method {0.__name__}.{1})'  # noqa: E501
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)

        for base in reversed(new_cls.__mro__):  # 多重継承を確認 !コマンドを登録
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]  # commandsから削除
                if elem in listeners:
                    del listeners[elem]  # listenersから削除
                is_static_method = isinstance(value, staticmethod)

                if isinstance(value, Command):
                    commands[elem] = value

                if is_static_method:  # staticmethodか確認
                    value = value.__func__  #
                # 関数をvalueに !valueが重要
                if isinstance(value, _BaseCommand):
                    commands[elem] = value
                elif inspect.iscoroutinefunction(value):
                    try:
                        value.__cog_listener__
                    except AttributeError:
                        continue
                    else:
                        if elem.startswith(('cog', 'bot_')):
                            raise TypeError(no_bot_cog.format(base, elem))
                        listeners[elem] = value
        new_cls.__cog_commands__ = list(commands.values())

        listeners_as_list: List[tuple[str, Any]] = []
        for listener in listeners.values():
            for listener_name in listener.__cog_listener_names__:
                listeners_as_list.append((listener_name, listener))

        new_cls.__cog_listeners__ = listeners_as_list
        return new_cls

    def __init__(self, *args: Tuple[Any], **kwargs: Dict[str, Any]):
        super().__init__(*args, **kwargs)

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__


class Cog(metaclass=CogMeta):
    __cog_name__ = ClassVar[str]
    __cog_settings__: Dict[str, Any] = {}
    __cog_listeners__: ClassVar[List[Tuple[str, str]]]
    __cog_commands__: List[Command] = []

    def __new__(cls, *args: Any, **kwargs: Any):
        self = super().__new__(cls)
        self.__cog_commands__ = tuple(cls.__cog_commands__)
        return self

    @classmethod
    def listener(cls, name: Optional[str] = None):
        def decorator(func: Cog):
            actual = func
            if isinstance(actual, staticmethod):
                actual = actual.__func__
            if not inspect.iscoroutinefunction(actual):
                raise TypeError(
                    'Listener function must be a coroutine function.'
                )
            actual.__cog_listener__ = True
            to_assign = name or actual.__name__
            try:
                actual.__cog_listener_names__.append(to_assign)
            except AttributeError:
                actual.__cog_listener_names__ = [to_assign]
            return func

        return decorator

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__

    def _inject(self, bot: BotBase):
        for command in self.__cog_commands__:
            bot.add_command(command, self.__cog_name__)

        for name, method_func in self.__cog_listeners__:
            bot.add_listener(getattr(self, name), name)

        return self



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.
    - Simplification of some functions.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import importlib
import inspect
import logging
import re
import sys
import traceback
from typing import Any, Callable, Coroutine, Dict, Optional, Tuple, Union

from aiohttp import ClientWebSocketResponse
from mipac.client import Client as API
from mipac.manager.client import ClientManager
from mipac.models.user import UserDetailed

from mipa.exception import WebSocketNotConnected, WebSocketReconnect
from mipa.gateway import MisskeyWebSocket
from mipa.router import Router
from mipa.state import ConnectionState
from mipa.utils import LOGING_LEVEL_TYPE, setup_logging

_log = logging.getLogger()


class Client:
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        max_capture: int = 100,
        **options: Dict[Any, Any],
    ):
        super().__init__(**options)
        self.max_capture = max_capture
        self._router: Router
        self.url = None
        self.extra_events: Dict[str, Any] = {}
        self.special_events: Dict[str, Any] = {}
        self.token: Optional[str] = None
        self.origin_uri: Optional[str] = None
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.core: API
        self._connection: ConnectionState
        self.user: UserDetailed
        self.ws: Optional[MisskeyWebSocket] = None
        self.should_reconnect = True

    def _get_state(self, **options: Any) -> ConnectionState:
        return ConnectionState(
            dispatch=self.dispatch, loop=self.loop, client=self
        )

    async def on_ready(self, ws: ClientWebSocketResponse):
        """
        on_readyのデフォルト処理

        Parameters
        ----------
        ws : WebSocketClientProtocol
        """

    def event(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_event(func, name)
            return func

        return decorator

    def add_event(
        self, func: Coroutine[Any, Any, Any], name: Optional[str] = None
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.special_events[name].append(func)
        else:
            self.special_events[name] = [func]

    def listen(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_listener(func, name)
            return func

        return decorator

    def add_listener(
        self,
        func: Union[Coroutine[Any, Any, Any], Callable[..., Any]],
        name: Optional[str] = None,
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')
        _log.debug(f'add_listener: {name} {func.__name__}')
        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def event_dispatch(
        self, event_name: str, *args: Tuple[Any], **kwargs: Dict[Any, Any]
    ) -> bool:
        """
        on_ready等といった

        Parameters
        ----------
        event_name :
        args :
        kwargs :

        Returns
        -------

        """

        ev = f'on_{event_name}'
        for event in self.special_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)
        return ev in dir(self)

    def dispatch(
        self, event_name: str, *args: tuple[Any], **kwargs: Dict[Any, Any]
    ):
        ev = f'on_{event_name}'
        for event in self.extra_events.get(ev, []):
            if inspect.ismethod(event):
                coro = event
                event = event.__name__
            else:
                foo = importlib.import_module(event.__module__)
                coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)

    def schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: tuple[Any],
        **kwargs: Dict[Any, Any],
    ) -> asyncio.Task[Any]:
        return self.loop.create_task(
            self._run_event(coro, event_name, *args, **kwargs),
            name=f'MiPA: {event_name}',
        )

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.__on_error(event_name)
            except asyncio.CancelledError:
                pass

    @staticmethod
    async def __on_error(event_method: str) -> None:
        print(f'Ignoring exception in {event_method}', file=sys.stderr)
        traceback.print_exc()

    async def on_error(self, err):
        self.event_dispatch('error', err)

    async def create_api_session(self, token: str, url: str) -> API:
        self.core = API(url, token)
        return self.core

    async def setup_hook(self) -> None:
        ...

    async def login(self, token: str, url: str):
        """
        ユーザーにログインし、ユーザー情報を取得します

        Parameters
        ----------
        token : str
            BOTにするユーザーのTOKEN
        url : str
            BOTにするユーザーがいるインスタンスのURL
        """

        core = await self.create_api_session(token, url)
        await core.http.login()
        self.user = await core.api.get_me()
        await self.setup_hook()

    async def _connect(
        self, *, timeout: int = 60, event_name: str = 'ready',
    ) -> None:
        self._connection = self._get_state()
        coro = MisskeyWebSocket.from_client(
            self, timeout=timeout, event_name=event_name
        )
        self.ws = await asyncio.wait_for(coro, timeout=60)
        while True:
            await self.ws.poll_event()

    async def connect(
        self, *, reconnect: bool = True, timeout: int = 60,
    ) -> None:
        self.should_reconnect = reconnect
        event_name = 'ready'
        while True:
            try:
                await self._connect(timeout=timeout, event_name=event_name)
            except (WebSocketReconnect, asyncio.exceptions.TimeoutError):
                if not self.should_reconnect:
                    break
                event_name = 'reconnect'
                await asyncio.sleep(3)

    async def disconnect(self):
        if not self.ws:
            raise WebSocketNotConnected()
        self.should_reconnect = False
        await self.ws.socket.close()

    @property
    def client(self) -> ClientManager:
        return self.core.api

    @property
    def router(self) -> Router:
        return self._router

    async def start(
        self,
        url: str,
        token: str,
        *,
        debug: bool = False,
        reconnect: bool = True,
        timeout: int = 60,
        is_ayuskey: bool = False,
        log_level: LOGING_LEVEL_TYPE = 'INFO',
    ):
        """
        Starting Bot

        Parameters
        ----------
        url: str
            Misskey Instance Websocket URL (wss://example.com)
        token: str
            User Token
        debug: bool, default False
            debugging mode
        reconnect: bool, default True
            coming soon...
        timeout: int, default 60
            Time until websocket times out
        """
        setup_logging(level=log_level)
        self.token = token
        url = url[:-1] if url[-1] == '/' else url
        split_url = url.split('/')

        if origin_url := re.search(r'wss?://(.*)', url):
            origin_url = (
                origin_url.group(0)
                .replace('wss', 'https')
                .replace('ws', 'http')
                .replace('/streaming', '')
            )
        else:
            origin_url = url
        if 'streaming' not in split_url:
            split_url.append('streaming')
            url = '/'.join(split_url)
        self.url = url.replace('https', 'wss').replace('http', 'ws')
        self.origin_url = origin_url
        await self.login(token, origin_url)
        await self.connect(reconnect=reconnect, timeout=timeout)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.
    - Simplification of some functions.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import asyncio
import inspect
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict

from mipac.models import Note
from mipac.models.chat import ChatMessage
from mipac.models.emoji import CustomEmoji
from mipac.models.note import NoteDeleted
from mipac.models.notification import (
    NotificationAchievement,
    NotificationFollow,
    NotificationFollowRequest,
    NotificationNote,
    NotificationPollEnd,
    NotificationReaction,
)
from mipac.models.reaction import PartialReaction
from mipac.models.user import UserDetailed
from mipac.types import INote
from mipac.types.chat import IChatMessage
from mipac.types.note import (
    INoteUpdated,
    INoteUpdatedDelete,
    INoteUpdatedReaction,
)
from mipac.util import str_lower, upper_to_lower

if TYPE_CHECKING:
    from mipac.types.notification import INotification
    from mipac.types.user import IUserDetailed

    from mipa.client import Client

_log = logging.getLogger(__name__)


class ConnectionState:
    def __init__(
        self,
        dispatch: Callable[..., Any],
        loop: asyncio.AbstractEventLoop,
        client: Client,
    ):
        self.__client: Client = client
        self.__dispatch = dispatch
        self.api = client.core.api
        self.loop: asyncio.AbstractEventLoop = loop
        self.parsers = parsers = {}
        for attr, func in inspect.getmembers(self):
            if attr.startswith('parse'):
                parsers[attr[6:].upper()] = func

    async def parse_emoji_added(self, message: Dict[str, Any]):
        self.__dispatch(
            'emoji_add', CustomEmoji(message['body']['emoji'], client=self.api)
        )

    async def parse_channel(self, message: Dict[str, Any]) -> None:
        """parse_channel is a function to parse channel event

        チャンネルタイプのデータを解析後適切なパーサーに移動させます

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        base_msg = upper_to_lower(message['body'])
        channel_type = str_lower(base_msg.get('type'))
        _log.debug(f'ChannelType: {channel_type}')
        _log.debug(f'recv event type: {channel_type}')
        if func := getattr(self, f'parse_{channel_type}', None):
            await func(base_msg['body'])
        else:
            _log.debug(f'Unknown event type: {channel_type}')

    async def parse_follow(self, message: IUserDetailed) -> None:
        """
        When you follow someone, this event will be called
        """
        user: UserDetailed = UserDetailed(
            message, client=self.api,
        )
        self.__dispatch('user_follow', user)

    async def parse_unfollow(self, message: IUserDetailed):
        """
        When you unfollow someone, this event will be called
        """
        user: UserDetailed = UserDetailed(
            message, client=self.api,
        )
        self.__dispatch('user_unfollow', user)

    async def parse_signin(self, message: Dict[str, Any]):
        """
        ログインが発生した際のイベント
        """

    async def parse_note_updated(self, message: INoteUpdated[Any]):
        message: Dict[str, Any] = upper_to_lower(message)
        if func := getattr(self, f'parse_{message["body"]["type"]}', None):
            await func(message)
        else:
            _log.debug(
                f'Unknown note_updated event type: {message["body"]["type"]}'
            )

    async def parse_deleted(self, note: INoteUpdated[INoteUpdatedDelete]):
        self.__dispatch('note_deleted', NoteDeleted(note))

    async def parse_unreacted(
        self, reaction: INoteUpdated[INoteUpdatedReaction]
    ):
        self.__dispatch('unreacted', PartialReaction(reaction))

    async def parse_reacted(
        self, reaction: INoteUpdated[INoteUpdatedReaction]
    ):
        self.__dispatch('reacted', PartialReaction(reaction))

    async def parse_me_updated(self, user: IUserDetailed):
        self.__dispatch('me_updated', UserDetailed(user, client=self.api))

    async def parse_read_all_announcements(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO: 実装

    async def parse_drive_file_created(self, message: Dict[str, Any]) -> None:
        self.__dispatch('drive_file_created', message)

    async def parse_read_all_unread_mentions(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_read_all_unread_specified_notes(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_read_all_channels(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    async def parse_read_all_notifications(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_url_upload_finished(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    async def parse_unread_mention(self, message: Dict[str, Any]) -> None:
        pass

    async def parse_unread_specified_note(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    async def parse_read_all_messaging_messages(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    async def parse_messaging_message(self, message: IChatMessage) -> None:
        """
        チャットが来た際のデータを処理する関数
        """
        self.__dispatch(
            'chat', ChatMessage(message, client=self.api),
        )

    async def parse_unread_messaging_message(
        self, message: IChatMessage
    ) -> None:
        """
        チャットが既読になっていない場合のデータを処理する関数
        """
        self.__dispatch(
            'chat_unread_message', ChatMessage(message, client=self.api),
        )

    async def parse_notification(self, message: Dict[str, Any]) -> None:
        """
        Parse notification event

        Parameters
        ----------
        message: Dict[str, Any]
            Received message
        """
        message: INotification = upper_to_lower(message)
        notification_map: dict[
            str,
            tuple[
                str,
                [
                    NotificationFollow
                    | NotificationNote
                    | NotificationReaction
                    | NotificationPollEnd
                    | NotificationFollowRequest
                ],
            ],
        ] = {
            'follow': ('user_followed', NotificationFollow),
            'mention': ('mention', NotificationNote),
            'reply': ('reply', NotificationNote),
            'renote': ('renote', NotificationNote),
            'quote': ('quote', NotificationNote),
            'reaction': ('reaction', NotificationReaction),
            'poll_vote': ('poll_vote', NotificationNote),
            'poll_ended': ('poll_end', NotificationPollEnd),
            'receive_follow_request': (
                'follow_request',
                NotificationFollowRequest,
            ),
            'follow_request_accepted': (
                'follow_request_accept',
                NotificationFollow,
            ),
            'achievement_earned': (
                'achievement_earned',
                NotificationAchievement,
            ),
        }
        dispatch_path, parse_class = notification_map.get(
            str_lower(message['type']), (None, None)
        )
        if dispatch_path:
            self.__dispatch(
                dispatch_path, parse_class(message, client=self.api)
            )

    async def parse_unread_notification(self, message: Dict[str, Any]) -> None:
        """
        未読の通知を解析する関数

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        # notification_type = str_lower(message['type'])
        # getattr(self, f'parse_{notification_type}')(message)

    async def parse_note(self, message: INote) -> None:
        """
        ノートイベントを解析する関数
        """
        note = Note(message, self.api)
        await self.__client.router.capture_message(note.id)
        self.__dispatch('note', note)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, List, Optional

from mipa.ext.commands._types import _BaseCommand

if TYPE_CHECKING:
    from mipa.ext.commands import Context


def hooked_wrapped_callback(coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        ret = await coro(*args, **kwargs)
        return ret

    return wrapped


class CMD:
    def __init__(
        self, cmd_type: str, key: str, func: 'Command', cog_name: str
    ):
        self.cmd_type = cmd_type
        self.key = key
        self.func = func
        self.cog_name = cog_name


class CommandManager:
    def __init__(self, *args, **kwargs):
        self.all_commands: List[CMD] = []
        super().__init__(*args, **kwargs)  # Clientクラスを初期化する

    def add_command(self, command: 'Command', cog_name: str):
        if not isinstance(command, Command):
            raise TypeError(f'{command}はCommandクラスである必要があります')
        command_type = 'regex' if command.regex else 'text'
        command_key = command.regex or command.text
        self.all_commands.append(
            CMD(command_type, command_key, command, cog_name)
        )


class Command(_BaseCommand):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self

    def __init__(self, func, regex: str, text: str, **kwargs):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'{func}はコルーチンでなければなりません')
        self.regex: str = regex
        self.text: str = text
        self.callback = func
        self.cog = None

    @property
    def qualified_name(self) -> str:
        return self.regex or self.text

    def __str__(self):
        return self.qualified_name

    @staticmethod
    async def _parse_arguments(ctx: Context):
        args = (ctx,) if ctx.cog is None else (ctx.cog, ctx)
        ctx.args = args + ctx.args
        return ctx

    async def invoke(self, ctx: Context, *args, **kwargs):
        ctx = await self._parse_arguments(ctx)
        await self.callback(*ctx.args, **ctx.kwargs)


def mention_command(regex: Optional[str] = None, text: Optional[str] = None):
    def decorator(func, **kwargs):
        return Command(func, regex=regex, text=text, **kwargs)

    return decorator
